import express from 'express'
import bodyParser from 'body-parser'
import cors from 'cors'
import { gql } from 'apollo-server'
import { buildSubgraphSchema } from '@apollo/federation'
import { applyMiddleware } from 'graphql-middleware'
import { ApolloServer } from 'apollo-server-express'
import { createServer } from 'http'
import { WebSocketServer } from 'ws'
import { ApolloServerPluginDrainHttpServer } from '@apollo/server/plugin/drainHttpServer'
import { useServer } from 'graphql-ws/lib/use/ws'
import { PubSub } from 'graphql-subscriptions'

export const pubSub = new PubSub()
export const initializeServer = async (
  models,
  typeDefs,
  resolvers,
  middlewares
) => {
  
  const corsOptions = { credentials: true, origin: true }
  const app = express()
  app.use(express.json())
  app.use(cors(corsOptions))
  app.use(bodyParser.urlencoded({ extended: false }))
  app.use(bodyParser.json())

  const schemaWithoutMiddleware = buildSubgraphSchema([
    {
      typeDefs: gql`
        ${typeDefs}
      `,
      resolvers: { ...resolvers },
      resolverValidationOptions: { requireResolversForResolveType: false },
    },
  ])

  const schema = applyMiddleware(schemaWithoutMiddleware, ...middlewares)

  const httpServer = createServer(app)
  const wsServer = new WebSocketServer({
    server: httpServer,
    path: '/graphql',
  })

  const getDynamicContext = async (ctx, msg, args) => {
    if (ctx.connectionParams.Authorization) {
      const userData = await validateToken(ctx.connectionParams.Authorization)
      return userData
    }
  }

  const serverCleanup = useServer(
    {
      schema,
      context: async (ctx, msg, args) => {
        return getDynamicContext(ctx, msg, args)
      },
    },
    wsServer
  )

  const server = new ApolloServer({
    schema,
    plugins: [
      ApolloServerPluginDrainHttpServer({ httpServer }),
      {
        async serverWillStart() {
          return {
            async drainServer() {
              await serverCleanup.dispose()
            },
          }
        },
      },
    ],
    context: async ({ req }) => {
      return { models, pubSub, req }
    },
  })

  await server.start()
  server.applyMiddleware({ app, cors: false })

  models.sequelize.sync({}).then(() => {
    httpServer.listen(process.env.PORT, () => {
      console.log(
        `🚀 Server ready at http://localhost:${process.env.PORT}${server.graphqlPath}`
      )
    })
  })
}
